from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from docugenr8.font import Font

from collections.abc import Callable

from .fragment import Fragment
from .word import Word


# CREATE WORDS
def create_words(
    unicode_text: str,
    current_font: Font,
    current_font_size: float,
    current_font_color: tuple[float, float, float],
    current_page_dummy: str | None = None,
    total_pages_dummy: str | None = None,
    page_num_dummy_length: int | None = None,
    page_number_presentation: Callable[[int], str] | None = None,
) -> list[Word]:
    char_idx = 0
    words: list[Word] = []
    while char_idx < len(unicode_text):
        (fragment, increment) = generate_fragment_with_increment(
            unicode_text,
            char_idx,
            current_font,
            current_font_size,
            current_font_color,
            current_page_dummy,
            total_pages_dummy,
            page_num_dummy_length,
            page_number_presentation,
        )
        if fragment._chars in {"\n", "\t", " "}:
            word = Word()
            word._add_fragment(fragment)
            words.append(word)
        else:
            if len(words) == 0:
                words.append(Word())
            if words[-1]._is_extendable:
                words[-1]._add_fragment(fragment)
            else:
                word = Word()
                word._add_fragment(fragment)
                words.append(word)
        char_idx += increment
    return words


def generate_fragment_with_increment(
    unicode_text: str,
    char_idx: int,
    current_font: Font,
    current_font_size: float,
    current_font_color: tuple[float, float, float],
    current_page_dummy: str | None,
    total_pages_dummy: str | None,
    page_num_dummy_length: int | None,
    page_number_presentation: Callable[[int], str] | None,
) -> tuple[Fragment, int]:
    if is_carriage_return_with_new_line(
        unicode_text,
        char_idx,
    ):
        fragment = generate_new_line_fragment(
            current_font,
            current_font_size,
            current_font_color,
        )
        return (fragment, 2)
    if is_only_carriage_return(unicode_text, char_idx):
        fragment = generate_new_line_fragment(
            current_font,
            current_font_size,
            current_font_color,
        )
        return (fragment, 1)
    if (
        current_page_dummy is not None
        and page_num_dummy_length is not None
        and page_number_presentation is not None
        and is_string_current_page_dummy(unicode_text, char_idx, current_page_dummy)
    ):
        (fragment, increment) = generate_current_page_fragment(
            current_font,
            current_font_size,
            current_font_color,
            current_page_dummy,
            page_num_dummy_length,
            page_number_presentation,
        )
        return (fragment, increment)
    if (
        total_pages_dummy is not None
        and page_num_dummy_length is not None
        and page_number_presentation is not None
        and is_string_total_pages_dummy(unicode_text, char_idx, total_pages_dummy)
    ):
        (fragment, increment) = generate_total_pages_fragment(
            current_font,
            current_font_size,
            current_font_color,
            total_pages_dummy,
            page_num_dummy_length,
            page_number_presentation,
        )
        return (fragment, increment)
    fragment_height = current_font._get_line_height(current_font_size)
    fragment_width = current_font._get_char_width(unicode_text[char_idx], current_font_size)
    fragment_ascent = current_font._get_ascent(current_font_size)
    fragment = Fragment(
        fragment_height,
        fragment_width,
        fragment_ascent,
        unicode_text[char_idx],
        current_font.name,
        current_font_size,
        current_font_color,
    )
    increment = 1
    return (fragment, increment)


def is_carriage_return_with_new_line(
    unicode_text: str,
    char_idx: int,
) -> bool:
    carriage_return = unicode_text[char_idx]
    try:
        new_line = unicode_text[char_idx + 1]
    except IndexError:
        return False
    if carriage_return == "\r" and new_line == "\n":
        return True
    return False


def is_only_carriage_return(unicode_text: str, char_idx: int) -> bool:
    if unicode_text[char_idx] == "\r":
        return True
    return False


def is_string_current_page_dummy(
    unicode_text: str,
    char_idx: int,
    current_page_dummy: str,
) -> bool:
    start = char_idx
    end = start + len(current_page_dummy)
    if unicode_text[start:end] == current_page_dummy:
        return True
    return False


def is_string_total_pages_dummy(unicode_text: str, char_idx: int, total_pages_dummy: str) -> bool:
    start = char_idx
    end = start + len(total_pages_dummy)
    if unicode_text[start:end] == total_pages_dummy:
        return True
    return False


def generate_current_page_fragment(
    current_font: Font,
    current_font_size: float,
    current_font_color: tuple[float, float, float],
    current_page_dummy: str,
    page_num_dummy_length: int,
    page_number_presentation: Callable[[int], str] = (str),
) -> tuple[Fragment, int]:
    fragment_width = page_num_dummy_length * current_font._get_char_width("0", current_font_size)
    fragment_height = current_font._get_line_height(current_font_size)
    fragment_ascent = current_font._get_ascent(current_font_size)
    fragment = Fragment(
        fragment_height,
        fragment_width,
        fragment_ascent,
        current_page_dummy,
        current_font.name,
        current_font_size,
        current_font_color,
    )
    fragment._page_number_presentation = page_number_presentation
    fragment._is_current_page_dummy = True
    increment = len(current_page_dummy)
    return (fragment, increment)


def generate_total_pages_fragment(
    current_font: Font,
    current_font_size: float,
    current_font_color: tuple[float, float, float],
    total_pages_dummy: str,
    page_num_dummy_length: int,
    page_number_presentation: Callable[[int], str] = (str),
) -> tuple[Fragment, int]:
    fragment_width = page_num_dummy_length * current_font._get_char_width("0", current_font_size)
    fragment_height = current_font._get_line_height(current_font_size)
    fragment_ascent = current_font._get_ascent(current_font_size)
    fragment = Fragment(
        fragment_height,
        fragment_width,
        fragment_ascent,
        total_pages_dummy,
        current_font.name,
        current_font_size,
        current_font_color,
    )
    fragment._page_number_presentation = page_number_presentation
    fragment._is_total_pages_dummy = True
    increment = len(total_pages_dummy)
    return (fragment, increment)


def generate_new_line_fragment(
    current_font: Font,
    current_font_size: float,
    current_font_color: tuple[float, float, float],
) -> Fragment:
    fragment_width = 0.0
    fragment_height = current_font._get_line_height(current_font_size)
    fragment_ascent = current_font._get_ascent(current_font_size)
    return Fragment(
        fragment_height, fragment_width, fragment_ascent, "\n", current_font.name, current_font_size, current_font_color
    )



"""docugenr8 module.

This module provides functionality to create and manipulate documents and its objects. It includes the `Document` class,
which represents a document and its pages. It has functions to add new pages, create and add objects to pages and pack
information into DTOs for usage in other modules where the information will be rendered.

Example:
    # Create a new document
    doc = Document()

    # Add first page to the document
    doc.add_page(595.2, 842.04)

    # Add font to the document
    doc.add_font(font_name="calibri", path="./fonts/calibri.ttf")

    # Create a text area and fill it with a text
    ta = doc.create_textarea(
    x=50,
    y=50,
    width=180,
    height=140)
    ta.add_text("The quick brown fox jumped over the lazy dog. 1234567890.")


    # Add the text area to the first page
    doc.pages[0].add_content(ta)

    # Pack document information into a DTO
    dto = doc.build_dto()
"""

from .document import Document as Document



from docugenr8_shared.dto import Dto
from docugenr8_shared.dto import DtoFont
from docugenr8_shared.dto import DtoPage

from .build_dto import _generate_dto_text_area
from .font import Font
from .page import Page
from .settings import Settings
from .text_area import TextArea


class Document:
    def __init__(self) -> None:
        self.settings = Settings()
        self.pages: list[Page] = []
        self.fonts: dict[str, Font] = {}

    def add_page(
        self,
        width: float,
        height: float,
    ) -> None:
        page = Page(width, height)
        self.pages.append(page)

    def add_font(
        self,
        font_name: str,
        path: str,
    ) -> None:
        font = Font(font_name, path)
        self.fonts[font_name] = font
        self.settings.font_current = font_name

    def create_textarea(self, x: float, y: float, width: float, height: float) -> TextArea:
        return TextArea(x, y, width, height, self)

    def _build_dto(self) -> Dto:
        dto = Dto()
        for font_name, font in self.fonts.items():
            dto_font = DtoFont(font_name, font.raw_data)
            dto.fonts.append(dto_font)
        for page_number, page in enumerate(self.pages):
            dto_page = DtoPage(page._width, page._height)
            dto.pages.append(dto_page)
            for content in page._contents:
                match content:
                    case TextArea():
                        content._build_current_page_fragments(page_number + 1)
                        content._build_total_pages_fragments(len(self.pages))
                        dto_page.contents.append(_generate_dto_text_area(content))
                    case _:
                        raise TypeError("Invalid content type.")
        return dto



from docugenr8_core.text_area.textarea import create_words

def test__check_parameters_of_one_line_text(doc_with_fonts):
    ta = doc_with_fonts.create_textarea(0, 0, 100, 100)
    ta.add_text("aaaaa aaaaa")
    assert len(ta._paragraphs) == 1
    assert ta._paragraphs[0]._textlines[0]._available_width == 45
    assert ta._paragraphs[0]._textlines[0]._height == 10
    assert len(ta._paragraphs[0]._textlines[0]._height_dict) == 1
    assert ta._paragraphs[0]._textlines[0]._ascent == 6
    assert len(ta._paragraphs[0]._textlines[0]._ascent_dict) == 1
    assert ta._paragraphs[0]._textlines[0]._leading == 0
    assert ta._paragraphs[0]._textlines[0]._width == 100
    assert len(ta._paragraphs[0]._textlines[0]._words) == 3
    assert ta._paragraphs[0]._textlines[0]._words[0]._chars == "aaaaa"
    assert ta._paragraphs[0]._textlines[0]._words[0]._width == 25
    assert ta._paragraphs[0]._textlines[0]._words[0]._chars == "aaaaa"
    assert ta._paragraphs[0]._textlines[0]._words[1]._chars == " "
    assert ta._paragraphs[0]._textlines[0]._words[2]._chars == "aaaaa"


def test__check_textline_available_width_with_spaces(doc_with_fonts):
    ta = doc_with_fonts.create_textarea(0, 0, 100, 100)
    ta.add_text("    aa  aa     ")
    assert ta._paragraphs[0]._textlines[0]._available_width == 50

def test__check_number_of_paragraphs(doc_with_fonts):
    ta = doc_with_fonts.create_textarea(0, 0, 100, 100)
    ta.add_text("aaaaa\naaaaa")
    assert len(ta._paragraphs) == 2
    assert ta._paragraphs[0]._textlines[0]._get_chars() == "aaaaa\n"
    assert ta._paragraphs[1]._textlines[0]._get_chars() == "aaaaa"

def test__spaces_at_the_end_of_textline(doc_with_fonts):
    ta = doc_with_fonts.create_textarea(0, 0, 100, 100)
    ta.add_text("    aaaa aaaa aaaa      aaaa    aaaa  aaaaaa    ")
    assert ta._paragraphs[0]._textlines[0]._available_width == 10
    assert ta._paragraphs[0]._textlines[0]._get_chars() == "    aaaa aaaa aaaa      "
    assert ta._paragraphs[0]._textlines[1]._available_width == 0
    assert ta._paragraphs[0]._textlines[1]._get_chars() == "aaaa    aaaa  aaaaaa    "

def test__tab_widths(doc_with_fonts):
    doc_with_fonts.settings.text_tab_size = 20
    ta = doc_with_fonts.create_textarea(0, 0, 100, 100)
    ta.add_text(" \taa\taa")
    assert ta._paragraphs[0]._textlines[0]._available_width == 50
    assert ta._paragraphs[0]._textlines[0]._words[1]._chars == "\t"
    assert ta._paragraphs[0]._textlines[0]._words[1]._width == 15
    new_words = create_words(
        "a",
        doc_with_fonts.fonts[doc_with_fonts.settings.font_current],
        doc_with_fonts.settings.font_size,
        doc_with_fonts.settings.font_color,
        doc_with_fonts.settings.page_num_current_page_dummy,
        doc_with_fonts.settings.page_num_total_pages_dummy,
        doc_with_fonts.settings.page_num_dummy_length,
        doc_with_fonts.settings.page_num_presentation,
        )
    ta._paragraphs[0]._textlines[0]._append_word(new_words[0], 1)
    assert ta._paragraphs[0]._textlines[0]._available_width == 50
    assert ta._paragraphs[0]._textlines[0]._words[2]._chars == "\t"
    assert ta._paragraphs[0]._textlines[0]._words[2]._width == 10

def test__merge_words(doc_with_fonts):
    ta = doc_with_fonts.create_textarea(0, 0, 100, 100)
    ta.add_text("aaaa")
    ta.add_text("aaaa")
    assert len(ta._paragraphs[0]._textlines[0]._words) == 1
    assert ta._paragraphs[0]._textlines[0]._words[0]._width == 40
    assert ta._paragraphs[0]._textlines[0]._words[0]._chars == "aaaaaaaa"
    assert ta._paragraphs[0]._textlines[0]._words[0]._textline == ta._paragraphs[0]._textlines[0]
    ta1 = doc_with_fonts.create_textarea(0, 0, 100, 100)
    current_page_dummy = doc_with_fonts.settings.page_num_current_page_dummy
    doc_with_fonts.settings.page_num_dummy_length = 2
    ta1.add_text(current_page_dummy)
    ta1.add_text("aa")
    ta1.add_text(current_page_dummy)
    assert len(ta1._paragraphs[0]._textlines[0]._words) == 1
    assert len(ta1._words_with_current_page_fragments) == 1
    assert ta1._words_with_current_page_fragments[0] == ta1._paragraphs[0]._textlines[0]._words[0]
    assert ta1._paragraphs[0]._textlines[0]._words[0]._width == 30



from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .fragment import Fragment
    from .paragraph import Paragraph
    from .textline import TextLine
    from .word import Word


class Word:
    def __init__(self) -> None:
        self._textline: None | TextLine = None
        self._fragments: deque[Fragment] = deque()
        self._chars = ""
        self._width = 0.0
        self._ascent = 0.0
        self._ascent_dict: dict[float, int] = {}
        self._height = 0.0
        self._height_dict: dict[float, int] = {}
        self._justify_space = 0.0
        self._is_extendable: bool = True
        self._should_render: bool = True
        self._current_page_fragments: list[Fragment] = []
        self._total_pages_fragments: list[Fragment] = []

    def _has_current_page_fragments(self) -> bool:
        if len(self._current_page_fragments) > 0:
            return True
        return False

    def _has_total_pages_fragments(self) -> bool:
        if len(self._total_pages_fragments) > 0:
            return True
        return False

    def _is_first_word_in_paragraph(self) -> bool:
        if self._textline is None:
            return False
        if self._textline._paragraph is None:
            return False
        first_linked_paragraph = self._textline._paragraph._get_first_linked_paragraph()
        return self == first_linked_paragraph._textlines[0]._words[0]

    def _is_last_word_in_paragraph(self) -> bool:
        if self._textline is None:
            return False
        if self._textline._paragraph is None:
            return False
        last_linked_paragraph = self._textline._paragraph._get_last_linked_paragraph()
        return self == last_linked_paragraph._textlines[-1]._words[-1]

    def _get_paragraph(self) -> Paragraph | None:
        if self._textline is None:
            return None
        return self._textline._paragraph

    def _remove_from_line(self) -> None:
        if self._textline is None:
            raise ValueError("Missing textline.")
        word_index = self._textline._words.index(self)
        self._textline._pop_word(word_index)

    def _add_page_number_to_textarea(self) -> None:
        if self._textline is None:
            return
        words_with_current_page_fragments = self._textline._paragraph._textarea._words_with_current_page_fragments
        words_with_total_pages_fragments = self._textline._paragraph._textarea._words_with_total_pages_fragments
        if self._has_current_page_fragments():
            words_with_current_page_fragments.append(self)
        if self._has_total_pages_fragments():
            words_with_total_pages_fragments.append(self)

    def _remove_page_number_from_textarea(self) -> None:
        if self._textline is None:
            return
        words_with_current_page_fragments = self._textline._paragraph._textarea._words_with_current_page_fragments
        words_with_total_pages_fragments = self._textline._paragraph._textarea._words_with_total_pages_fragments
        if self in words_with_current_page_fragments:
            words_with_current_page_fragments.remove(self)
        if self in words_with_total_pages_fragments:
            words_with_total_pages_fragments.remove(self)

    def _get_origin_width(self) -> float:
        width = 0.0
        for fragment in self._fragments:
            width += fragment._width
        return width

    def _append_height(self, height: float) -> None:
        if height not in self._height_dict:
            self._height_dict[height] = 1
        else:
            self._height_dict[height] += 1
        if height > self._height:
            height_diff = height - self._height
            self._height += height_diff
            if self._textline is not None:
                self._textline._append_height(self._height)

    def _append_ascent(self, ascent: float) -> None:
        if ascent not in self._ascent_dict:
            self._ascent_dict[ascent] = 1
        else:
            self._ascent_dict[ascent] += 1
        if ascent > self._ascent:
            ascent_diff = ascent - self._ascent
            self._ascent += ascent_diff
            if self._textline is not None:
                self._textline._append_ascent(self._ascent)

    def _remove_height(self, height: float) -> None:
        self._height_dict[height] -= 1
        if self._height_dict[height] > 0:
            return
        del self._height_dict[height]
        if len(self._height_dict) == 0:
            self._height = 0.0
            return
        if self._height == height:
            self._height = max(self._height_dict)
            if self._textline is not None:
                raise NotImplementedError("Needs implementation " "for removing height from " "a textline.")
            # height_diff = self.height - removed_height

    def _remove_ascent(self, ascent: float) -> None:
        self._ascent_dict[ascent] -= 1
        if self._ascent_dict[ascent] > 0:
            return
        removed_ascent = self._ascent_dict.pop(ascent, None)
        if len(self._ascent_dict) == 0:
            self._ascent = 0.0
            return
        if self._ascent == removed_ascent:
            self._ascent = max(self._ascent_dict)
            if self._textline is not None:
                raise NotImplementedError("Needs implementation " "for removing ascent from " "a textline.")
            # ascent_diff = self.ascent - removed_ascent

    def _add_fragment(self, fragment: Fragment) -> Word:
        fragment._word = self
        if fragment._chars in {" ", "\t", "\n"}:
            self._is_extendable = False
        if fragment._chars in {"\t", "\n"}:
            self._should_render = False
        self._fragments.append(fragment)
        self._adjust_width_by_difference(fragment._width)
        self._append_height(fragment._height)
        self._append_ascent(fragment._ascent)
        self._chars += fragment._chars
        if fragment._is_current_page_dummy:
            self._current_page_fragments.append(fragment)
        if fragment._is_total_pages_dummy:
            self._total_pages_fragments.append(fragment)
        return self

    def _pop_fragment_left(self) -> Fragment:
        first_fragment = self._fragments.popleft()
        first_fragment._word = None
        self._chars = self._chars[len(first_fragment._chars) :]
        self._width -= first_fragment._width
        self._remove_height(first_fragment._height)
        self._remove_ascent(first_fragment._ascent)
        return first_fragment

    def _pop_fragment(self, index: int) -> Fragment:
        if index < 0 or index > len(self._fragments):
            raise IndexError(f"Fragment index {index} is out of range in word.")
        fragment_to_remove = self._fragments[index]
        fragment_to_remove._word = None
        chars_left = self._chars[:index]
        chars_right = self._chars[index + 1 :]
        self._chars = chars_left + chars_right
        self._fragments.remove(fragment_to_remove)
        self._width -= fragment_to_remove._width
        self._remove_height(fragment_to_remove._height)
        self._remove_ascent(fragment_to_remove._ascent)
        if fragment_to_remove._is_current_page_dummy:
            self._current_page_fragments.remove(fragment_to_remove)
        if fragment_to_remove._is_total_pages_dummy:
            self._total_pages_fragments.remove(fragment_to_remove)
        return fragment_to_remove

    def _adjust_width_by_difference(self, width_diff: float) -> None:
        self._width += width_diff

    def _calc_width_with_justify(self) -> float:
        return self._width + self._justify_space

    def _is_from_textarea(
        self,
    ) -> bool:
        if self._get_paragraph() is not None:
            return True
        return False

    def _is_from_buffer(self) -> bool:
        if self._get_paragraph() is None:
            return True
        return False



from collections.abc import Callable


class Settings:
    def __init__(self) -> None:
        self.font_current: None | str = None
        self.font_size: float = 11.0
        self.font_color: tuple[int, int, int] = (0, 0, 0)
        self.text_tab_size = 35.4375
        self.text_h_align: str = "left"
        self.text_v_align: str = "top"
        self.textline_height_ratio: float = 1.2
        self.text_split_words: bool = True
        self.paragraph_first_line_indent: float = 20.0
        self.paragraph_hanging_indent: float = 0.0
        self.paragraph_left_indent: float = 10.0
        self.paragraph_right_indent: float = 10.0
        self.paragraph_space_before: float = 10.0
        self.paragraph_space_after: float = 10.0
        self.page_num_current_page_dummy: str = "%%pn%%"
        self.page_num_total_pages_dummy: str = "%%tp%%"
        self.page_num_dummy_length: int = 2
        self.page_num_presentation: Callable[[int], str] = str



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Callable

    from .word import Word


class Fragment:
    def __init__(
        self,
        height: float,
        width: float,
        ascent: float,
        chars: str,
        font_name: str,
        font_size: float,
        font_color: tuple[int, int, int],
    ) -> None:
        self._height = height
        self._width = width
        self._ascent = ascent
        self._chars = chars
        self._font_name = font_name
        self._font_size = font_size
        self._font_color = font_color
        self._word: None | Word = None
        self._page_number_presentation: Callable[[int], str] | None = None
        self._is_current_page_dummy: bool = False
        self._is_total_pages_dummy: bool = False

    def _adjust_width(self, new_width: float) -> None:
        width_diff = new_width - self._width
        self._width += width_diff
        if self._word is not None:
            self._word._adjust_width_by_difference(width_diff)



"""build_dto module.

This module provides functionality to build Data Transfer Objects (DTOs) for
transferring document information between modules.
"""

from docugenr8_shared.dto import DtoFragment
from docugenr8_shared.dto import DtoParagraph
from docugenr8_shared.dto import DtoTextArea
from docugenr8_shared.dto import DtoTextLine
from docugenr8_shared.dto import DtoWord

from .text_area.fragment import Fragment
from .text_area.paragraph import Paragraph
from .text_area.textarea import TextArea
from .text_area.textline import TextLine
from .text_area.word import Word


def _generate_dto_fragment(
    x: float,
    y: float,
    dto_word: DtoWord,
    fragment: Fragment,
) -> DtoFragment:
    """Generate Dto fragment object. Dto fragment object is the smallest posible object that holds one character.

    Args:
        x (float): x position of the fragment
        y (float): y position of the fragment
        dto_word (DtoWord): Parent Dto word object
        fragment (Fragment): Word fragment from text area

    Returns:
        DtoFragment: Dto fragment object
    """
    dto_fragment = DtoFragment(x, y, dto_word)
    dto_fragment.width = fragment._width
    dto_fragment.height = fragment._height
    dto_fragment.baseline = y + fragment._ascent
    dto_fragment.chars = fragment._chars
    dto_fragment.font_name = fragment._font_name
    dto_fragment.font_size = fragment._font_size
    dto_fragment.font_color = fragment._font_color
    return dto_fragment


def _generate_dto_word(
    x: float,
    y: float,
    dto_textline: DtoTextLine,
    word: Word,
) -> DtoWord:
    """_summary_.

    Args:
        x (float): _description_
        y (float): _description_
        dto_textline (DtoTextLine): _description_
        word (Word): _description_

    Returns:
        DtoWord: _description_
    """
    dto_word = DtoWord(x, y, dto_textline)
    dto_word.width = word._width
    dto_word.height = word._height
    dto_word.baseline = y + word._ascent
    dto_word.justify_space = word._justify_space
    for fragment in word._fragments:
        dto_fragment = _generate_dto_fragment(x, y + (word._ascent - fragment._ascent), dto_word, fragment)
        dto_word.fragments.append(dto_fragment)
        dto_word.textline.fragments.append(dto_fragment)
        dto_word.textline.paragraph.fragments.append(dto_fragment)
        dto_word.textline.paragraph.text_area.fragments.append(dto_fragment)
        x += dto_fragment.width
    return dto_word


def _generate_dto_textline(
    x: float, y: float, dto_paragraph: DtoParagraph, textline: TextLine, justify_padding_after: float
) -> DtoTextLine:
    dto_textline = DtoTextLine(x, y, dto_paragraph, justify_padding_after)
    dto_textline.width = textline._width
    dto_textline.height = textline._height
    dto_textline.baseline = y + textline._ascent
    dto_textline.space_after = textline._leading
    dto_textline.paragraph_h_align = dto_paragraph.h_align
    x_offset = 0.0
    justify_space = 0.0
    match dto_textline.paragraph_h_align:
        case "left":
            pass
        case "right":
            x_offset = textline._available_width
        case "center":
            x_offset = (textline._available_width) / 2
        case "justify":
            justify_space = _calculate_justify_space(textline)
    x += x_offset
    for word in textline._words:
        dto_word = _generate_dto_word(x, y + (textline._ascent - word._ascent), dto_textline, word)
        dto_textline.words.append(dto_word)
        if word._chars == " ":
            x += dto_word.width + justify_space
        else:
            x += dto_word.width
    return dto_textline


def _generate_dto_paragraph(
    x: float,
    y: float,
    paragraph: Paragraph,
    dto_text_area: DtoTextArea,
    should_distrubute_space_in_lines: bool,
    height_diff: float,
    last_paragraph: bool,
) -> DtoParagraph:
    dto_paragraph = DtoParagraph(x, y, dto_text_area)
    dto_paragraph.width = paragraph._width
    dto_paragraph.height = paragraph._height
    dto_paragraph.chars = paragraph._get_chars()
    dto_paragraph.tab_size = paragraph._tab_size
    dto_paragraph.line_height_ratio = paragraph._line_height_ratio
    dto_paragraph.first_line_indent = paragraph._first_line_indent
    dto_paragraph.hanging_indent = paragraph._hanging_indent
    dto_paragraph.left_indent = paragraph._left_indent
    dto_paragraph.right_indent = paragraph._right_indent
    dto_paragraph.space_before = paragraph._space_before
    dto_paragraph.space_after = paragraph._space_after
    dto_paragraph.h_align = paragraph._h_align
    num_of_lines_in_paragraph = len(paragraph._textlines)
    y += dto_paragraph.space_before
    for line in paragraph._textlines:
        line_justify_padding = 0.0
        if should_distrubute_space_in_lines:
            if not last_paragraph:
                line_justify_padding = height_diff / num_of_lines_in_paragraph
            else:
                line_justify_padding = height_diff / (num_of_lines_in_paragraph - 1)
        if line._is_first_line_in_paragraph():
            x_offset = x + dto_paragraph.left_indent + dto_paragraph.first_line_indent
        else:
            x_offset = x + dto_paragraph.left_indent + dto_paragraph.hanging_indent
        dto_textline = _generate_dto_textline(
            x_offset,
            y,
            dto_paragraph,
            line,
            line_justify_padding,
        )
        dto_paragraph.textlines.append(dto_textline)
        y += dto_textline.height
        y += dto_textline.space_after
        y += dto_textline.justify_padding_after
    y += dto_paragraph.space_after
    dto_paragraph.height += height_diff
    return dto_paragraph


def _generate_dto_text_area(text_area: TextArea) -> DtoTextArea:
    y_offset: float = 0.0
    between_paragraphs_padding: float = 0.0
    should_distrubute_space_in_lines: bool = False
    num_of_lines_in_text_area: int = 0
    # ******************************
    match text_area._v_align:
        case "top":
            pass
        case "bottom":
            y_offset += text_area._available_height
        case "center":
            y_offset += text_area._available_height / 2
        case "justify-paragraphs":
            between_paragraphs_padding = text_area._available_height / (len(text_area._paragraphs) - 1)
        case "justify-lines":
            should_distrubute_space_in_lines = True
            for paragraph in text_area._paragraphs:
                num_of_lines_in_text_area += len(paragraph._textlines)
        case _:
            raise ValueError("Invalid value for vertical alignment.")
    # ******************************
    dto_text_area = DtoTextArea(text_area._x, text_area._y, text_area._width, text_area._height)
    dto_text_area.height_empty_space = text_area._available_height
    dto_text_area.v_align = text_area._v_align

    y = text_area._y + y_offset
    for paragraph in text_area._paragraphs:
        height_diff = 0.0
        last_paragraph = False
        if should_distrubute_space_in_lines:
            if paragraph != text_area._paragraphs[-1]:
                height_diff = (len(paragraph._textlines) / num_of_lines_in_text_area) * text_area._available_height
            else:
                height_diff = (
                    (len(paragraph._textlines) - 1) / num_of_lines_in_text_area
                ) * text_area._available_height
                last_paragraph = True
        # ******************************
        dto_paragraph = _generate_dto_paragraph(
            text_area._x, y, paragraph, dto_text_area, should_distrubute_space_in_lines, height_diff, last_paragraph
        )
        dto_text_area.paragraphs.append(dto_paragraph)
        # ******************************
        y += dto_paragraph.height
        if paragraph != text_area._paragraphs[-1]:
            y += between_paragraphs_padding
    return dto_text_area


def _calculate_justify_space(textline: TextLine) -> float:
    if _textline_has_only_empty_spaces(textline):
        return 0
    if _textline_has_only_one_word(textline):
        return 0
    if textline._is_last_line_in_paragraph():
        return 0
    if textline._available_width == 0:
        return 0
    start_pos = _get_index_first_not_empty_space_word(textline)
    end_pos = _get_index_last_not_empty_space_word(textline)
    num_inner_words = 0
    for i in range(start_pos, end_pos):
        if textline._words[i]._chars == " ":
            num_inner_words += 1
    if num_inner_words == 0:
        return 0
    return textline._available_width / num_inner_words


def _textline_has_only_empty_spaces(textline: TextLine) -> bool:
    return all(word._chars in {" ", "\t", "\n"} for word in textline._words)


def _textline_has_only_one_word(textline: TextLine) -> bool:
    if len(textline._words) <= 1:
        return True
    return False


def _get_index_first_not_empty_space_word(textline: TextLine) -> int:
    for index, word in enumerate(textline._words):
        if word._chars not in {" ", "\t", "\n"}:
            return index
    raise ValueError("Could not obtain index for the first word that" " is not empty space.")


def _get_index_last_not_empty_space_word(textline: TextLine) -> int:
    for index, word in enumerate(reversed(textline._words)):
        if word._chars not in {" ", "\t", "\n"}:
            return len(textline._words) - index - 1
    raise ValueError("Could not obtain index for the last word that" " is not empty space.")



from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING

from .word import Word


if TYPE_CHECKING:
    from .paragraph import Paragraph


class TextLine:
    def __init__(
        self,
        paragraph: Paragraph,
    ) -> None:
        self._paragraph = paragraph
        self._width = 0.0
        self._available_width = 0.0
        self._set_textline_indent()
        self._spaces_width_at_the_end = 0.0
        self._spaces_at_the_end: list[Word] = []
        self._h_align = self._paragraph._h_align
        self._words: deque[Word] = deque()
        self._height: float = 0.0
        self._height_dict: dict[float, int] = {}
        self._ascent: float = 0.0
        self._ascent_dict: dict[float, int] = {}
        self._prev: None | TextLine = None
        self._next: None | TextLine = None
        self._tabs: list[Word] = []
        self._leading: float = 0.0
        self._inner_spaces: list[Word] = []

    def _append_word(self, word: Word, index: int | None = None) -> None:
        if index is None:
            index = len(self._words)
        if index < 0 or index > len(self._words):
            raise IndexError(f"Index {index} out of range.")
        word._textline = self
        self._words.insert(index, word)
        self._append_height(word._height)
        self._append_ascent(word._ascent)
        self._set_available_width(word)
        self._set_spaces_width_at_the_end()
        self._set_tab_width_if_needed(word)
        self._merge_words_if_possible(word)

    def _pop_word(self, index: int | None = None) -> Word:
        if index is None:
            index = len(self._words) - 1
        if index < 0 or index > len(self._words) - 1:
            raise IndexError(f"Index {index} out of range.")
        word_to_remove = self._words[index]
        self._words.remove(word_to_remove)
        self._remove_ascent(word_to_remove._ascent)
        self._remove_height(word_to_remove._height)
        self._set_spaces_width_at_the_end()
        self._available_width += word_to_remove._width
        if word_to_remove in self._tabs:
            self._tabs.remove(word_to_remove)
        for tab in self._tabs:
            self._recalculate_tab_width(tab)
        word_to_remove._textline = None
        if len(self._words) == 0:
            self._remove_line()
        if self._next is not None:
            self._set_leading()
        return word_to_remove

    def _set_spaces_width_at_the_end(self) -> None:
        if len(self._words) == 0:
            return
        for word in reversed(self._words):
            if word._chars != " ":
                break
            if word not in self._spaces_at_the_end:
                self._spaces_at_the_end.append(word)
                self._spaces_width_at_the_end += word._width
                self._available_width += word._width
        if self._words[-1]._chars != " ":
            self._spaces_at_the_end.clear()
            self._available_width -= self._spaces_width_at_the_end
            self._spaces_width_at_the_end = 0

    def _set_available_width(self, word: Word) -> None:
        if word._chars == "\t":
            return
        self._available_width -= word._width

    def _set_tab_width_if_needed(self, word: Word) -> None:
        if word._chars == "\t" and word not in self._tabs:
            self._add_tab(word)
        for tab in self._tabs:
            self._recalculate_tab_width(tab)

    def _add_tab(self, word: Word) -> None:
        if len(self._tabs) == 0:
            self._tabs.append(word)
            return
        word_index_in_words = self._words.index(word)
        for tab in self._tabs:
            tab_index_in_words = self._words.index(tab)
            if word_index_in_words < tab_index_in_words:
                tab_index_in_tabs = self._tabs.index(tab)
                self._tabs.insert(tab_index_in_tabs, word)
                return
        self._tabs.append(word)

    def _recalculate_tab_width(self, tab: Word) -> None:
        old_tab_width = tab._width
        tab_size = self._paragraph._tab_size
        width = 0
        for word in self._words:
            if word == tab:
                break
            width += word._width
        tab._width = tab_size - (width % tab_size)
        self._available_width = self._available_width + old_tab_width - tab._width

    def _test_tab_width(self) -> float:
        tab_size = self._paragraph._tab_size
        width = 0
        for word in self._words:
            width += word._width
        return tab_size - (width % tab_size)

    def _merge_words_if_possible(self, word: Word) -> None:
        if len(self._words) == 1:
            return
        if not word._is_extendable:
            return
        word_index = self._words.index(word)
        word_index_before = word_index - 1
        if word_index_before >= 0 and self._words[word_index_before]._is_extendable:
            self._merge_words(self._words[word_index_before], self._words[word_index])
            word_index -= 1
        word_index_after = word_index + 1
        if word_index_after < len(self._words) and self._words[word_index_after]._is_extendable:
            self._merge_words(self._words[word_index], self._words[word_index_after])

    def _merge_words(self, word_left: Word, word_right: Word) -> None:
        if not word_left._is_extendable or not word_right._is_extendable:
            return
        new_word = Word()
        for fragment in word_left._fragments:
            new_word._add_fragment(fragment)
        for fragment in word_right._fragments:
            new_word._add_fragment(fragment)
        index = self._words.index(word_left)
        self._words.insert(index, new_word)
        new_word._textline = word_left._textline
        new_word._add_page_number_to_textarea()
        textline = new_word._textline
        if textline is None:
            raise ValueError("Textline is not present.")
        textline._append_ascent(new_word._ascent)
        textline._append_height(new_word._height)

        textline._remove_ascent(word_left._ascent)
        textline._remove_ascent(word_right._ascent)
        textline._remove_height(word_left._height)
        textline._remove_height(word_right._height)

        word_left._remove_page_number_from_textarea()
        word_right._remove_page_number_from_textarea()
        self._words.remove(word_left)
        self._words.remove(word_right)

    # split word takes another paramether:
    # width to split the first part of the word
    def _split_word(
        self,
        right_word: Word,
        left_width_to_split: float,
    ) -> None:
        if right_word not in self._words:
            raise IndexError(f"Word {right_word._chars} " f"is not present in Textline {self._get_chars()}.")
        left_word = Word()
        while left_word._width + right_word._fragments[0]._width <= left_width_to_split:
            first_fragment = right_word._pop_fragment(0)
            left_word._add_fragment(first_fragment)
        if left_word._width == 0:
            return
        left_word._textline = self
        word_index = self._words.index(right_word)
        self._words.insert(word_index, left_word)
        right_word._remove_page_number_from_textarea()
        left_word._add_page_number_to_textarea()
        right_word._add_page_number_to_textarea()
        self._append_ascent(right_word._ascent)
        self._append_height(right_word._height)

    def _can_move_first_word_from_next_line(self) -> bool:
        first_word = self._paragraph._get_first_word_from_next_textline(self)
        if first_word is None:
            return False
        if first_word._chars in {" ", "\n"}:
            return True
        word_width = 0.0
        if first_word._chars == "\t":
            word_width = self._test_tab_width()
        else:
            word_width = first_word._width
        if (self._available_width - self._spaces_width_at_the_end - word_width) < 0:
            return False
        return True

    def _push_exceeded_words_to_next_textline(self) -> None:
        removed_words = deque()
        while self._available_width < 0:
            removed_words.appendleft(self._pop_word())
        if self._next is None:
            self._paragraph._create_textline()
            self._set_leading()
        if self._next is not None:
            while len(removed_words) > 0:
                self._next._append_word(removed_words.pop(), 0)

    def _remove_line_and_get_words(
        self,
    ) -> deque[Word]:
        removed_words = deque()
        while len(self._words) > 0:
            self._words[0]._remove_page_number_from_textarea()
            removed_word = self._pop_word(0)
            removed_words.append(removed_word)
        return removed_words

    def _is_first_line_in_paragraph(self) -> bool:
        first_linked_paragraph = self._paragraph._get_first_linked_paragraph()
        return self == first_linked_paragraph._textlines[0]

    def _is_last_line_in_paragraph(self) -> bool:
        last_linked_paragraph = self._paragraph._get_last_linked_paragraph()
        return self == last_linked_paragraph._textlines[-1]

    def _get_chars(self) -> str:
        chars = ""
        for word in self._words:
            chars += word._chars
        return chars

    def _adjust_words_between_textlines(self) -> None:
        if self._word_width_exceeds_texline_width():
            return
        if self._available_width < 0:
            self._push_exceeded_words_to_next_textline()
        self._pull_available_words_from_next_line()

    def _word_width_exceeds_texline_width(
        self,
    ) -> bool:
        if self._paragraph._textarea._document.settings.text_split_words and self._words[0]._width > self._width:
            self._split_word(self._words[0], self._width)
        if self._words[0]._width > self._width:
            textline_index = self._paragraph._textlines.index(self)
            paragraph_index = self._paragraph._textarea._paragraphs.index(self._paragraph)
            self._paragraph._textarea._empty_textlines_and_paragraphs_from_line(textline_index, paragraph_index)
            self._paragraph._textarea._state_textline_width_overflow = True
            return True
        return False

    def _pull_available_words_from_next_line(self) -> None:
        while self._can_move_first_word_from_next_line():
            if self._next is None:
                raise ValueError("Next textline is not present.")
            self._append_word(self._next._pop_word(0))

    def _append_height(self, height: float) -> None:
        if height not in self._height_dict:
            self._height_dict[height] = 1
        else:
            self._height_dict[height] += 1
        if height <= self._height:
            return
        height_diff = height - self._height
        self._height += height_diff
        if self._next is not None:
            self._set_leading()
        self._paragraph._change_height(height_diff)

    def _append_ascent(self, ascent: float) -> None:
        if ascent not in self._ascent_dict:
            self._ascent_dict[ascent] = 1
        else:
            self._ascent_dict[ascent] += 1
        if ascent > self._ascent:
            ascent_diff = ascent - self._ascent
            self._ascent += ascent_diff

    def _remove_height(self, height: float) -> None:
        self._height_dict[height] -= 1
        if self._height_dict[height] > 0:
            return
        del self._height_dict[height]
        if len(self._height_dict) == 0:
            # ???????
            return
        if self._height != height:
            return
        self._height = max(self._height_dict)
        height_diff = self._height - height
        if self._next is not None:
            self._set_leading()
        self._paragraph._change_height(height_diff)

    def _remove_ascent(self, ascent: float) -> None:
        self._ascent_dict[ascent] -= 1
        if self._ascent_dict[ascent] > 0:
            return
        del self._ascent_dict[ascent]
        if self._ascent != ascent:
            return
        if len(self._ascent_dict) == 0:
            self._ascent = 0.0
            return
        self._ascent = max(self._ascent_dict)

    def _remove_line(self) -> None:
        height_to_remove = self._height + self._leading
        if self._next is not None:
            self._next._prev = self._prev
        if self._prev is not None:
            self._prev._next = self._next
            self._prev._set_leading()
        self._leading = 0.0
        self._next = None
        self._prev = None
        paragraph = self._paragraph
        paragraph._textlines.remove(self)
        paragraph._change_height(-height_to_remove)
        if len(paragraph._textlines) == 0:
            textarea = paragraph._textarea
            textarea._available_height -= paragraph._space_before + paragraph._space_after
            textarea._paragraphs.remove(paragraph)

    def _calculate_leading(self) -> float:
        return (self._height * self._paragraph._line_height_ratio) - self._height

    def _set_leading(self) -> None:
        leading_diff = 0
        if self._next is None:
            leading_diff = -self._leading
            self._leading = 0
        else:
            leading_diff = self._calculate_leading() - self._leading
            self._leading += leading_diff
        self._paragraph._change_height(leading_diff)

    def _set_width_and_available_space(self, width: float) -> None:
        width_diff = width - self._width
        self._width += width_diff
        self._available_width += width_diff

    def _set_textline_indent(
        self,
    ) -> None:
        if self._paragraph._prev_linked_paragraph is None and len(self._paragraph._textlines) == 0:
            self._set_width_and_available_space(
                self._paragraph._width
                - self._paragraph._left_indent
                - self._paragraph._first_line_indent
                - self._paragraph._right_indent
            )
        else:
            self._set_width_and_available_space(
                self._paragraph._width
                - self._paragraph._left_indent
                - self._paragraph._hanging_indent
                - self._paragraph._right_indent
            )



from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from docugenr8.document import Document
    from docugenr8.font import Font

    from .textline import TextLine
    from .word import Word

from .paragraph import Paragraph
from .words_creation import create_words


class TextArea:
    def __init__(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        document: Document,
    ) -> None:
        self._x = x
        self._y = y
        self._width = width
        self._height = height
        self._document = document
        self._v_align = document.settings.text_v_align
        self._h_align = document.settings.text_h_align
        self._available_height = height
        self._paragraphs: deque[Paragraph] = deque()
        self._buffer: deque[Word] = deque()
        self._next_textarea: None | TextArea = None
        self._prev_textarea: None | TextArea = None
        self._state_accepts_words: bool = True
        self._state_textline_width_overflow: bool = False
        # late binding when adding text
        self._current_font: None | Font = None
        self._current_font_size: float = 0
        self._current_font_color: tuple[int, int, int] = (0, 0, 0)

        # words to fill in with page numbers
        self._words_with_current_page_fragments: list[Word] = []
        self._words_with_total_pages_fragments: list[Word] = []

    def add_text(
        self,
        unicode_text: str,
    ) -> None:
        self._save_font_attributes_from_document_settings()
        self._create_and_insert_words_into_buffer(unicode_text)
        self._distribute_words_in_all_areas()

    def _distribute_words_in_all_areas(self) -> None:
        textarea = self._get_the_first_textarea_to_accept_words()
        if textarea is None:
            return
        textarea._pull_and_push_words()

    def _save_font_attributes_from_document_settings(self) -> None:
        if self._document is not None:
            font_name = self._document.settings.font_current
            if font_name is None:
                raise TypeError("Current name of a font must be set.")
            self._current_font = self._document.fonts[font_name]
            self._current_font_size = self._document.settings.font_size
            self._current_font_color = self._document.settings.font_color
        if self._current_font is None:
            raise TypeError("The font object is missing.")

    def _create_and_insert_words_into_buffer(
        self,
        unicode_text: str,
    ) -> None:
        if self._current_font is None:
            raise ValueError("Current font must be set in settings.")
        words = create_words(
            unicode_text,
            self._current_font,
            self._current_font_size,
            self._current_font_color,
            self._document.settings.page_num_current_page_dummy,
            self._document.settings.page_num_total_pages_dummy,
            self._document.settings.page_num_dummy_length,
            self._document.settings.page_num_presentation,
        )
        buffer = self._get_buffer()
        buffer.extend(words)

    def _set_buffer(self, words: deque[Word]) -> None:
        current_textarea = self
        while current_textarea._next_textarea is not None:
            current_textarea = current_textarea._next_textarea
        current_textarea._buffer = words

    def _get_buffer(self) -> deque[Word]:
        current_textarea = self
        while current_textarea._next_textarea is not None:
            current_textarea = current_textarea._next_textarea
        return current_textarea._buffer

    def _get_the_last_word_from_textareas(self) -> Word | None:
        textarea = self._get_the_first_textarea_to_accept_words()
        if textarea is None or textarea._is_empty():
            return None
        return textarea._paragraphs[-1]._textlines[-1]._words[-1]

    def _get_the_first_textarea_to_accept_words(self) -> TextArea | None:
        current_textarea = self
        while current_textarea._state_accepts_words is False:
            if current_textarea._next_textarea is not None:
                current_textarea = current_textarea._next_textarea
            else:
                return None
        return current_textarea

    def _distrubute_words_in_all_paragraphs(self) -> None:
        for paragraph in self._paragraphs:
            paragraph._adjust_words_between_textlines()

    def _pull_and_push_words(self) -> None:
        if self._available_height >= 0:
            self._pull_next_available_word()
        if self._available_height < 0:
            self._state_accepts_words = False
            self._push_excess_words_to_next_available_place()

    def _pull_next_available_word(self) -> None:
        while self._available_height >= 0:
            next_word = self._get_next_word()
            if next_word is None:
                return
            if next_word._is_from_textarea():
                self._create_paragraph_from_textarea_if_needed(next_word)
                next_word._remove_page_number_from_textarea()
                next_word._remove_from_line()
            if next_word._is_from_buffer():
                self._create_paragraph_from_buffer_if_needed()
                self._get_buffer().remove(next_word)
            self._append_word_to_textarea(next_word)
            if self._state_textline_width_overflow:
                self._state_textline_width_overflow = False
                break

    def _append_word_to_textarea(self, word: Word) -> None:
        if word._has_current_page_fragments():
            self._words_with_current_page_fragments.append(word)
        if word._has_total_pages_fragments():
            self._words_with_total_pages_fragments.append(word)
        self._paragraphs[-1]._append_word_right(word)

    def _push_excess_words_to_next_available_place(self) -> deque[Word] | None:
        textline_to_push = self._paragraphs[-1]._textlines[-1]
        while self._available_height < 0:
            words_to_push = textline_to_push._remove_line_and_get_words()
            if self._next_textarea is not None:
                self._next_textarea._generate_paragraph_when_pushing_words(textline_to_push)
                self._next_textarea._add_words_front_to_textarea(words_to_push)
            else:
                words_to_push.extend(self._get_buffer())
                self._set_buffer(words_to_push)
            if len(self._paragraphs) == 0:
                return
            textline_to_push = self._paragraphs[-1]._textlines[-1]
        if self._next_textarea is not None:
            if len(self._paragraphs) == 0:
                self._next_textarea._empty_textarea()
            else:
                self._next_textarea._pull_and_push_words()

    def _create_paragraph_from_buffer_if_needed(
        self,
    ) -> None:
        if len(self._paragraphs) == 0:
            new_paragraph = Paragraph(self)
            self._paragraphs.append(new_paragraph)
            if self._prev_textarea is not None:
                prev_paragraph = self._prev_textarea._paragraphs[-1]
                prev_paragraph._next_linked_paragraph = new_paragraph
                new_paragraph._prev_linked_paragraph = prev_paragraph
                new_paragraph._copy_paragraph_parameters_from(prev_paragraph)
        if self._paragraphs[-1]._ends_with_br:
            self._paragraphs.append(Paragraph(self))

    def _create_paragraph_from_textarea_if_needed(
        self,
        next_word: Word,
    ) -> None:
        paragraph = next_word._get_paragraph()
        if paragraph is None:
            raise TypeError("Next word is missing paragraph object.")
        if next_word._is_first_word_in_paragraph():
            if next_word._textline is not None:
                paragraph._set_non_first_line_indent(next_word._textline)
            new_paragraph = Paragraph(self)
            self._paragraphs.append(new_paragraph)
            new_paragraph._copy_paragraph_parameters_from(paragraph)
            new_paragraph._next_linked_paragraph = paragraph
            paragraph._prev_linked_paragraph = new_paragraph
            return
        if next_word._is_last_word_in_paragraph():
            self._paragraphs[-1]._next_linked_paragraph = None
            paragraph._prev_linked_paragraph = None
            return
        if self._paragraphs[-1]._next_linked_paragraph != paragraph:
            self._paragraphs[-1]._next_linked_paragraph = paragraph
            paragraph._prev_linked_paragraph = self._paragraphs[-1]
            return

    def _generate_paragraph_when_pushing_words(self, textline_to_push: TextLine) -> None:
        if len(self._paragraphs) == 0 and textline_to_push._is_first_line_in_paragraph():
            new_paragraph = Paragraph(self)
            self._paragraphs.append(new_paragraph)
            if textline_to_push._paragraph is not None:
                textline_to_push._paragraph._next_linked_paragraph = None
            return
        if len(self._paragraphs) == 0 and not textline_to_push._is_first_line_in_paragraph():
            new_paragraph = Paragraph(self)
            self._paragraphs.append(new_paragraph)
            if textline_to_push._paragraph is not None:
                new_paragraph._copy_paragraph_parameters_from(textline_to_push._paragraph)
            new_paragraph._prev_linked_paragraph = textline_to_push._paragraph
            if textline_to_push._paragraph is not None:
                textline_to_push._paragraph._next_linked_paragraph = new_paragraph
            return
        if textline_to_push._is_last_line_in_paragraph():
            new_paragraph = Paragraph(self)
            self._paragraphs.appendleft(new_paragraph)
            if textline_to_push._paragraph is not None:
                new_paragraph._copy_paragraph_parameters_from(textline_to_push._paragraph)
            new_paragraph._prev_linked_paragraph = textline_to_push._paragraph
            if textline_to_push._paragraph is not None:
                textline_to_push._paragraph._next_linked_paragraph = new_paragraph
        if textline_to_push._is_first_line_in_paragraph():
            self._paragraphs[0]._prev_linked_paragraph = None
            if textline_to_push._paragraph is not None:
                textline_to_push._paragraph._next_linked_paragraph = None

    def _add_words_front_to_textarea(self, words: deque[Word]) -> None:
        while len(words) > 0:
            word = words.popleft()
            self._paragraphs[0]._textlines[0]._append_word(word, 0)
        self._paragraphs[0]._adjust_words_between_textlines()

    def _get_next_word(self) -> Word | None:
        """gets the next word from either next paragraph if there is one,
        or from the words buffer
        """
        textarea = self._get_next_textarea_with_words()
        if textarea is not None:
            return textarea._paragraphs[0]._textlines[0]._words[0]
        buffer = self._get_buffer()
        if len(buffer) == 0:
            return None
        return buffer[0]

    def _get_next_textarea_with_words(self) -> TextArea | None:
        textarea = self._next_textarea
        while textarea is not None:
            if not textarea._is_empty():
                return textarea
            textarea = textarea._next_textarea
        return None

    def _is_empty(self) -> bool:
        if len(self._paragraphs) == 0:
            return True
        if len(self._paragraphs[0]._textlines) == 0:
            return True
        if len(self._paragraphs[0]._textlines[0]._words) == 0:
            return True
        return False

    def _build_current_page_fragments(self, current_page: int) -> None:
        for word in self._words_with_current_page_fragments:
            for fragment in word._current_page_fragments:
                if fragment._page_number_presentation is None:
                    continue
                font = self._document.fonts[fragment._font_name]
                chars = fragment._page_number_presentation(current_page)
                page_number_width = 0.0
                if len(chars) > 0:
                    fragment._chars = chars
                    for char in chars:
                        page_number_width += font._get_char_width(char, fragment._font_size)
                fragment._adjust_width(page_number_width)
                if (
                    fragment._word is not None
                    and fragment._word._textline is not None
                    and fragment._word._textline._paragraph is not None
                    and fragment._word._textline._paragraph._textarea is not None
                ):
                    paragraph = fragment._word._textline._paragraph
                    paragraph._adjust_words_between_textlines()
                    textarea = fragment._word._textline._paragraph._textarea
                    textarea._pull_and_push_words()

    def _build_total_pages_fragments(self, total_pages: int) -> None:
        for word in self._words_with_total_pages_fragments:
            for fragment in word._total_pages_fragments:
                if fragment._page_number_presentation is None:
                    continue
                font = self._document.fonts[fragment._font_name]
                chars = fragment._page_number_presentation(total_pages)
                new_length = 0.0
                if len(chars) > 0:
                    fragment._chars = chars
                    for char in chars:
                        new_length += font._get_char_width(char, fragment._font_size)
                fragment._adjust_width(new_length)
                if (
                    fragment._word is not None
                    and fragment._word._textline is not None
                    and fragment._word._textline._paragraph is not None
                    and fragment._word._textline._paragraph._textarea is not None
                ):
                    paragraph = fragment._word._textline._paragraph
                    paragraph._adjust_words_between_textlines()
                    textarea = fragment._word._textline._paragraph._textarea
                    textarea._pull_and_push_words()

    def link_textarea(self, next_textarea: TextArea) -> None:
        last_textarea = self
        while last_textarea._next_textarea is not None:
            last_textarea = last_textarea._next_textarea
        last_textarea._next_textarea = next_textarea
        next_textarea._buffer = last_textarea._buffer
        last_textarea._buffer = deque()
        self._distribute_words_in_all_areas()

    def set_width(self, new_width: float) -> None:
        self._width = new_width
        for paragraph in self._paragraphs:
            paragraph._set_paragraph_width(new_width)
        self._distribute_words_in_all_areas()

    def _empty_textlines_and_paragraphs_from_line(self, textline_index: int, paragraph_index: int) -> None:
        removed_words = deque()
        first_paragraph_to_remove_words_from = self._paragraphs[paragraph_index]
        textline = first_paragraph_to_remove_words_from._textlines[textline_index]
        while textline_index < len(first_paragraph_to_remove_words_from._textlines):
            textline = self._paragraphs[paragraph_index]._textlines[textline_index]
            removed_words.extend(textline._remove_line_and_get_words())
        if len(self._paragraphs) > 0 and (first_paragraph_to_remove_words_from == self._paragraphs[paragraph_index]):
            paragraph_index += 1
        while paragraph_index < len(self._paragraphs):
            paragraph = self._paragraphs[paragraph_index]
            removed_words.extend(paragraph._remove_paragraph_from_text_area())
        textarea_to_empty = self._next_textarea
        while textarea_to_empty is not None:
            removed_words.extend(textarea_to_empty._empty_textarea())
            textarea_to_empty = textarea_to_empty._next_textarea
        word_buffer = self._get_buffer()
        removed_words.extend(word_buffer)
        self._set_buffer(removed_words)

    def _empty_textarea(self) -> deque[Word]:
        removed_words = deque()
        while len(self._paragraphs) > 0:
            paragraph = self._paragraphs[0]
            removed_words.extend(paragraph._remove_paragraph_from_text_area())
        return removed_words



from .text_area import TextArea


class Page:
    def __init__(self, width: float, height: float) -> None:
        self._width = width
        self._height = height
        self._contents: list[object] = []

    def add_content(self, content: object) -> None:
        match content:
            case TextArea():
                self._contents.append(content)
            case _:
                raise TypeError("Content type not defined.")



from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .textarea import TextArea
    from .word import Word

from .textline import TextLine


class Paragraph:
    def __init__(self, textarea: TextArea) -> None:
        self._textarea = textarea
        self._width: float = textarea._width
        self._height: float = 0.0
        self._space_before: float = textarea._document.settings.paragraph_space_before
        self._space_after: float = textarea._document.settings.paragraph_space_after
        self._change_height(self._space_before + self._space_after)
        self._line_height_ratio: float = textarea._document.settings.textline_height_ratio
        self._tab_size: float = textarea._document.settings.text_tab_size
        self._first_line_indent: float = textarea._document.settings.paragraph_first_line_indent
        self._hanging_indent: float = textarea._document.settings.paragraph_hanging_indent
        self._left_indent: float = textarea._document.settings.paragraph_left_indent
        self._right_indent: float = textarea._document.settings.paragraph_right_indent
        self._h_align: str = textarea._h_align
        self._textlines: list[TextLine] = []
        self._ends_with_br: bool = False
        self._next_linked_paragraph: None | Paragraph = None
        self._prev_linked_paragraph: None | Paragraph = None

    def _get_first_linked_paragraph(self) -> Paragraph:
        first_linked_paragraph = self
        while first_linked_paragraph._prev_linked_paragraph is not None:
            first_linked_paragraph = first_linked_paragraph._prev_linked_paragraph
        return first_linked_paragraph

    def _get_last_linked_paragraph(self) -> Paragraph:
        last_linked_paragraph = self
        while last_linked_paragraph._next_linked_paragraph is not None:
            last_linked_paragraph = last_linked_paragraph._next_linked_paragraph
        return last_linked_paragraph

    def _copy_paragraph_parameters_from(self, other_paragraph: Paragraph) -> None:
        self._line_height_ratio = other_paragraph._line_height_ratio
        self._tab_size = other_paragraph._tab_size
        self._first_line_indent = other_paragraph._first_line_indent
        self._hanging_indent = other_paragraph._hanging_indent
        self._left_indent = other_paragraph._left_indent
        self._right_indent = other_paragraph._right_indent
        height_diff = other_paragraph._space_before - self._space_before
        height_diff += other_paragraph._space_after - self._space_after
        self._space_before = other_paragraph._space_before
        self._space_after = other_paragraph._space_after
        self._change_height(height_diff)

    def _get_chars(self) -> str:
        chars = ""
        for line in self._textlines:
            chars += line._get_chars()
        return chars

    def _adjust_words_between_textlines(self) -> None:
        if len(self._textlines) == 0:
            return
        line = self._textlines[0]
        while line is not None and line._next is not None:
            line._adjust_words_between_textlines()
            line = line._next

    def _change_height(self, height_diff: float) -> None:
        self._height += height_diff
        self._textarea._available_height -= height_diff

    def _append_word_left(self, word: Word) -> None:
        if len(self._textlines) == 0:
            self._create_textline()
        if word._chars == "\n":
            self._ends_with_br = True
        self._textlines[0]._append_word(word, 0)

    def _append_word_right(self, word: Word) -> None:
        if len(self._textlines) == 0:
            self._create_textline()
        if word._chars == "\n":
            self._ends_with_br = True
        last_line = self._textlines[-1]
        last_line._append_word(word)
        last_line._adjust_words_between_textlines()

    def _create_textline(
        self,
        index: int | None = None,
    ) -> None:
        if index is None:
            index = len(self._textlines)
        if index < 0 or index > len(self._textlines):
            raise IndexError(f"Index {index} out of range in paragraph.")

        textline = TextLine(self)
        if self._prev_linked_paragraph is None and len(self._textlines) == 0:
            self._set_first_line_indent(textline)
        else:
            self._set_non_first_line_indent(textline)
        prev_textline = self._textlines[index - 1] if index > 0 else None
        next_textline = self._textlines[index + 1] if index < (len(self._textlines) - 1) else None
        self._textlines.insert(index, textline)
        textline._prev = prev_textline
        textline._next = next_textline
        if prev_textline is not None:
            prev_textline._next = textline
            prev_textline._set_leading()
        if next_textline is not None:
            next_textline._prev = textline

    def _set_non_first_line_indent(self, textline: TextLine) -> None:
        textline._set_width_and_available_space(
            self._width - self._left_indent - self._hanging_indent - self._right_indent
        )

    def _set_first_line_indent(self, textline: TextLine) -> None:
        textline._set_width_and_available_space(
            self._width - self._left_indent - self._first_line_indent - self._right_indent
        )

    def _pop_word_front_from_paragraph(self) -> Word:
        return self._textlines[0]._pop_word(0)

    def _pop_word_back_from_paragraph(self) -> Word:
        return self._textlines[-1]._pop_word(len(self._textlines[-1]._words) - 1)

    def _set_paragraph_width(self, new_width: float) -> None:
        width_diff = new_width - self._width
        self._width += width_diff
        for line in self._textlines:
            if line is self._textlines[0] and self._prev_linked_paragraph is None:
                self._set_first_line_indent(line)
            else:
                self._set_non_first_line_indent(line)
        self._adjust_words_between_textlines()

    def _get_first_word_from_next_textline(self, textline: TextLine) -> Word | None:
        if textline not in self._textlines:
            raise ValueError("Textline object is not present " "in parent paragraph object.")
        textline_index = self._textlines.index(textline)
        next_textline_index = textline_index + 1
        if next_textline_index > len(self._textlines):
            raise ValueError("Invalid index of the textline.")
        if next_textline_index == len(self._textlines):
            return None
        if len(self._textlines[next_textline_index]._words) == 0:
            return None
        return self._textlines[next_textline_index]._words[0]

    def _remove_paragraph_from_text_area(self) -> deque[Word]:
        removed_words = deque()
        while len(self._textlines) > 0:
            removed_words.extend(self._textlines[0]._remove_line_and_get_words())
        return removed_words



import os
import pathlib
from io import BytesIO

from fontTools import ttLib


TAB = 9
NEW_LINE = 10
SPACE = 32


class Font:
    def __init__(self, font_name: str, path: str) -> None:
        self.name = font_name
        self.raw_data = pathlib.Path(_resolve_file_path(path)).read_bytes()
        self.ttfont = ttLib.TTFont(BytesIO(self.raw_data), recalcTimestamp=False)
        self.cmap = self.ttfont.getBestCmap()
        self.em: float = self.ttfont["head"].unitsPerEm  # type: ignore
        self.ascent_per_em: float = self.ttfont["hhea"].ascent  # type: ignore
        self.descent_per_em: float = self.ttfont["hhea"].descent  # type: ignore
        self.line_height_per_em: float = self.ascent_per_em - self.descent_per_em

    def _get_char_width(self, char: str, font_size: float) -> float:
        unicode = ord(char)
        if unicode in {TAB, NEW_LINE}:
            return 0.0
        try:
            glyph_name = self.cmap[unicode]
            glyph_width = self.ttfont["hmtx"].metrics[glyph_name][0]  # type: ignore
        # for unicodes not defined in font
        except KeyError:
            glyph_name = ".notdef"
            glyph_width = self.ttfont["hmtx"].metrics[glyph_name][0]  # type: ignore
        return font_size * glyph_width / self.em

    def _get_ascent(self, font_size: float) -> float:
        return font_size * (self.ascent_per_em / self.em)

    def _get_descent(self, font_size: float) -> float:
        return font_size * (self.descent_per_em / self.em)

    def _get_line_height(self, font_size: float) -> float:
        return font_size * (self.line_height_per_em / self.em)


def _resolve_file_path(path: str) -> pathlib.Path:
    root_dir = pathlib.Path(os.getcwd())
    return root_dir / path
